package edu.fu_tdg.lodib.rdf_generator;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edu.fu_tdg.lodib.rdf_generator.distributions.Distribution;
import edu.fu_tdg.lodib.rdf_generator.distributions.UniformDistribution;
import edu.fu_tdg.lodib.rdf_generator.literal.DateGenerator;
import edu.fu_tdg.lodib.rdf_generator.literal.DoubleGenerator;
import edu.fu_tdg.lodib.rdf_generator.literal.LiteralGenerator;
import edu.fu_tdg.lodib.rdf_generator.literal.StringGenerator;
import edu.fu_tdg.lodib.rdf_generator.xml.ClassInstance;
import edu.fu_tdg.lodib.rdf_generator.xml.ConnectClass;
import edu.fu_tdg.lodib.rdf_generator.xml.ConnectLiteral;
import edu.fu_tdg.lodib.rdf_generator.xml.CreateClass;
import edu.fu_tdg.lodib.rdf_generator.xml.CreateLiteral;
import edu.fu_tdg.lodib.rdf_generator.xml.ForEach;
import edu.fu_tdg.lodib.rdf_generator.xml.Generator;
import edu.fu_tdg.lodib.rdf_generator.xml.Namespace;
import edu.fu_tdg.lodib.utils.Utils;

public class RDFGenerator {
	private Map<String, String> namespaces = new HashMap<String, String>();
	private Map<String, String> instancePrefixes = new HashMap<String, String>();
	private Map<String, Long> classesCount = new HashMap<String, Long>();
	private BufferedWriter output = null;
	
	public void generateRDF(File xml, File outputRDF) throws RDFGeneratorException {
		try {
			// Create output buffer
			output = new BufferedWriter(new FileWriter(outputRDF));
			
			// Transform XML file into internal objects
			edu.fu_tdg.lodib.rdf_generator.xml.RDFGenerator xmlObject = 
					Utils.unmarshall(edu.fu_tdg.lodib.rdf_generator.xml.RDFGenerator.class, new FileInputStream(xml));
			
			// Add namespaces to map
			for (Namespace nm : xmlObject.getNamespace())
				namespaces.put(nm.getPrefix(), nm.getValue());
			
			// Add instance prefixes
			for (ClassInstance ci : xmlObject.getClassInstance())
				instancePrefixes.put(ci.getClazz(), ci.getPrefix());
			
			// Parse create
			for (CreateClass cr : xmlObject.getCreateClass())
				parseCreateClass(cr);
			
			// Parse for each
			for (ForEach fe : xmlObject.getForEach())
				parseForEach(fe);
			
			output.close();
		} catch (Exception oops) {
			throw new RDFGeneratorException("Error when generating RDF: " + oops, oops);
		}
	}
	
	private CreateClassResult parseCreateClass(CreateClass cr) throws IOException {
		// Get class name
		String className = cr.getClazz();
		
		// Get number of instances to generate
		Long number = getNumberOfInstancesOrValues(cr.getExactValue(), cr.getMinRange(), cr.getMaxRange());
		
		// Update class map, count range before and after
		Long init = Long.valueOf(0);
		if (classesCount.containsKey(className))
			init = classesCount.get(className);
		updateClassesCount(classesCount, className, number);
		Long end = classesCount.get(className);
		CreateClassResult ret = new CreateClassResult(init, end, className);
					
		// Add triples to output
		for (long i = init; i < end; i++)
			Utils.addTriple(output, Utils.toURI(instancePrefixes.get(className) + i, namespaces), Utils.toURI(Utils.RDF_TYPE, namespaces), Utils.toURI(className, namespaces));
		
		return ret;
	}
	
	private CreateLiteralResult parseCreateLiteral(CreateLiteral cr) throws IOException, ParseException {
		// Get type name
		String typeName = cr.getType();
		
		// Get number of literals to generate
		Long number = getNumberOfInstancesOrValues(cr.getExactValue(), cr.getMinRange(), cr.getMaxRange());
		// Parse literal generator
		LiteralGenerator gen = parseGenerator(cr.getGenerator(), typeName);
		// Generate literals
		List<String> literals = new ArrayList<String>();
		if (gen != null)
			for (long i = 0; i < number; i++)
				literals.add(gen.generateValue());
		
		return new CreateLiteralResult(literals, typeName);
	}
	
	private LiteralGenerator parseGenerator(Generator g, String type) throws ParseException {
		LiteralGenerator gen = null;
		
		// Create string generator
		if (g.getType().equals(LiteralGenerator.STRING_GENERATOR))
			gen = new StringGenerator(Integer.valueOf(g.getMinRange()), Integer.valueOf(g.getMaxRange()), type, g.getLangTag(), g.getValues(), namespaces);
		else if (g.getType().equals(LiteralGenerator.DOUBLE_GENERATOR))
			gen = new DoubleGenerator(Double.valueOf(g.getMinRange()), Double.valueOf(g.getMaxRange()), type, namespaces);
		else if (g.getType().equals(LiteralGenerator.DATE_GENERATOR))
			gen = new DateGenerator(Utils.DATE_FORMAT.parse(g.getMinRange()), Utils.DATE_FORMAT.parse(g.getMaxRange()), type, namespaces);
		// TODO 1: Add more literal types
		
		return gen;
	}
	
	private Long getNumberOfInstancesOrValues(String exactValue, String minRange, String maxRange) {
		// Get number of instances/literals to generate
		Long number = null;
		if (exactValue != null)
			// Get exact number
			number = Long.valueOf(exactValue);
		else {
			// Compute range and randomly choose one
			Long initialNumber = Long.valueOf(minRange);
			Long finalNumber = Long.valueOf(maxRange);
			
			// TODO 2: This distribution should be configurable
			Distribution<Long> dist = new UniformDistribution<Long>(initialNumber, finalNumber + 1);
			number = dist.getValue();
		}
		
		return number;
	}
	
	private void updateClassesCount(Map<String, Long> classesCount, String className, Long number) {
		// Add zero if class is new
		if (!classesCount.containsKey(className))
			classesCount.put(className, Long.valueOf(0));
		// Update number of classes
		classesCount.put(className, classesCount.get(className) + number);
	}
	
	private void parseForEach(ForEach fe) throws IOException, InstantiationException, IllegalAccessException, ClassNotFoundException, ParseException {
		// Get class name
		String className = fe.getClazz();
		
		// Iterate over each instance of the current class
		for (long i = 0; i < classesCount.get(className); i++) {
			String iClassInstance = instancePrefixes.get(className) + i;
			
			// Parse create class
			CreateClassResult retClass = null;
			if (fe.getCreateClass() != null)
				retClass = parseCreateClass(fe.getCreateClass());
			
			// Parse create literal
			CreateLiteralResult retLiteral = null;
			if (fe.getCreateLiteral() != null)
				retLiteral = parseCreateLiteral(fe.getCreateLiteral());
			
			// For each connectClass
			for (ConnectClass c : fe.getConnectClass()) {
				if (retClass == null) {
					// No create, so we expect a distribution
					Distribution<Long> distrib = null;
					// TODO 2: Add more distributions!
					if (c.getUniform() != null) {
						Long init = Long.valueOf(c.getUniform().getInit());
						Long end = Long.valueOf(c.getUniform().getEnd());
						
						distrib = new UniformDistribution<Long>(init, end + 1);
					}
					long v = distrib.getValue() - 1;
					String subject = (c.getSubject().equals(className) ? iClassInstance : instancePrefixes.get(c.getSubject()) + v), 
							object = (c.getObject().equals(className) ? iClassInstance : instancePrefixes.get(c.getObject()) + v);
					Utils.addTriple(output, Utils.toURI(subject, namespaces), Utils.toURI(c.getPredicate(), namespaces), Utils.toURI(object, namespaces));
				} else {
					for (long j = retClass.getInit(); j < retClass.getEnd(); j++) {
						String jClassInstance = instancePrefixes.get(retClass.getClassName()) + j;
						
						String subject = (c.getSubject().equals(className) ? iClassInstance : jClassInstance), 
								object = (c.getObject().equals(className) ? iClassInstance : jClassInstance);
						
						Utils.addTriple(output, Utils.toURI(subject, namespaces), Utils.toURI(c.getPredicate(), namespaces), Utils.toURI(object, namespaces));
					}
				}
			}
			
			// For each connectLiteral
			for (ConnectLiteral c : fe.getConnectLiteral())
				// We expect retLiteral
				for (String s : retLiteral.getLiterals())
					Utils.addTriple(output, Utils.toURI(iClassInstance, namespaces), Utils.toURI(c.getPredicate(), namespaces), s);
		}
	}
}
